__author__ = 'Michael Wilson & Lindo Nkambule'

import hailtop.batch as hb
import hail as hl
import pandas as pd
from gwaspy.utils.get_file_size import bytes_to_gb
from gwaspy.phasing.get_filebase import get_vcf_filebase
from typing import List
from typing import Union


def concat_vcfs(b: hb.batch.Batch,
                vcf_basename: str = None,
                vcfs_to_merge: List = None,
                output_type: str = 'bcf',
                software: str = None,
                chrom: str = None,
                docker_img: str = 'docker.io/lindonkambule/gwaspy:v1',
                cpu: int = 8,
                out_dir: str = None):

    global index_cmd

    out_type = 'b' if output_type == 'bcf' else 'z'
    threads = cpu - 1
    vcfs_sizes_sum = 0
    merge_vcf_i = ''

    out_filename = f'{vcf_basename}.{chrom}.phased.{software}.bcf' if output_type == 'bcf' else \
        f'{vcf_basename}.{chrom}.phased.{software}.vcf.gz'
    out_index_name = f'{vcf_basename}.{chrom}.phased.{software}.bcf.csi' if output_type == 'bcf' else \
        f'{vcf_basename}.{chrom}.phased.{software}.vcf.gz.csi'

    for line in vcfs_to_merge:
        vcfs_sizes_sum += 1 + bytes_to_gb(line)

    mem = 'highmem' if vcfs_sizes_sum > 2 else 'standard'
    disk_size = 10 + vcfs_sizes_sum

    concat = b.new_job(name=f'concat-{vcf_basename}')
    concat.memory(mem)
    concat.storage(f'{disk_size}Gi')
    concat.image(docker_img)
    concat.cpu(cpu)

    for line in vcfs_to_merge:
        input_vcf = b.read_input_group(vcf=line,
                                       ind=f'{line}.csi')
        merge_vcf_i += f'{input_vcf.vcf} '

    cmd = f'''
        bcftools concat \
            --no-version \
            --output-type {out_type} \
            --output {out_filename} \
            --threads {threads} \
            --ligate \
            {merge_vcf_i}
    '''

    concat.command(cmd)
    # index the merged output
    concat.command(f'bcftools index {out_filename}')

    concat.command(f'mv {out_filename} {concat.ofile}')
    concat.command(f'mv {out_index_name} {concat.idx}')
    b.write_output(concat.ofile, f'{out_dir}/GWASpy/{vcf_basename}/Phasing/phased_merged/{out_filename}')
    b.write_output(concat.idx, f'{out_dir}/GWASpy/{vcf_basename}/Phasing/phased_merged/{out_index_name}')


def run_concat(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
               input_vcf: str = None,
               output_type: str = 'bcf',
               reference: str = 'GRCh38',
               software: str = None,
               out_dir: str = None):

    print(f'\n3. CONCAT {input_vcf}\n')
    vcf_filebase = get_vcf_filebase(input_vcf)

    # get the regions so we can map each file to its specific region
    regions = pd.read_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/regions.lines', sep='\t', names=['reg', 'ind'])
    regions_dict = pd.Series(regions.reg.values, index=regions.ind).to_dict()

    concat_b = hb.Batch(backend=backend, name=f'concat-phased-chunks-{vcf_filebase}')

    if software == 'shapeit':
        phased_vcf_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_scatter/*.shapeit.bcf')
    else:
        phased_vcf_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_scatter/*.eagle.bcf')

    for i in range(1, 24):
        if reference == 'GRCh38':
            if i == 23:
                chrom = 'chrX'
            else:
                chrom = f'chr{i}'

            out_chrom_name = chrom
        else:
            chrom = str(i)
            out_chrom_name = f'chr{chrom}'

        chrom_phased_files_to_concat = []

        for file in phased_vcf_chunks:
            f = file['path']
            vcf_basename = get_vcf_filebase(f)
            file_index = int(vcf_basename.split('.')[-3])
            file_region = regions_dict[file_index]
            map_chrom = file_region.split(':')[0]
            if map_chrom == chrom:
                chrom_phased_files_to_concat.append(f)

        # naturally sort the list of files to merge
        from gwaspy.utils.natural_sort import natural_keys
        chrom_phased_files_to_concat.sort(key=natural_keys)

        # checkpoint to see if file already exists to avoid redoing things
        chrom_out_filename = f'{vcf_filebase}.{out_chrom_name}.phased.{software}.bcf' if output_type == 'bcf' else \
            f'{vcf_filebase}.{out_chrom_name}.phased.{software}.vcf.gz'
        chrom_out_filname_path = f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_merged/{chrom_out_filename}'

        if hl.hadoop_exists(chrom_out_filname_path):
            continue
        else:
            concat_vcfs(b=concat_b, vcfs_to_merge=chrom_phased_files_to_concat, vcf_basename=vcf_filebase,
                        output_type=output_type, software=software, chrom=out_chrom_name, out_dir=out_dir)

    concat_b.run()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS'
]

setup(name='gwaspy',
      version='0.1.1',
      author='Lindokuhle Nkambule',
      author_email='lnkambul@broadinstitute.org',
      url='https://gwaspy.readthedocs.io/',
      project_urls={"GitHub": "https://github.com/atgu/GWASpy"},
      description='GWASpy: A Python package for performing GWAS QC, PCA, phasing, and genotype imputation.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='MIT',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'preimp_qc = gwaspy.preimp_qc.preimp_qc:main',
              'pca = gwaspy.pca.pca:main',
              'imputation = gwaspy.imputation.imputation:main',
              'phasing = gwaspy.phasing.phasing:main'
          ]
      },
      classifiers=classifiers,
      keywords='',
      install_requires=['hail', 'matplotlib', 'numpy', 'pandas', 'pylatex', 'plotly'],
      zip_safe=False
      )


__author__ = 'Lindo Nkambule'

import argparse
import hail as hl


def pca(
        ref_dirname: str = 'gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/',
        ref_basename: str = 'unrelated',
        ref_info: str = 'gs://hgdp-1kg/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_sample_info.unrelateds.pca_outliers_removed.with_project.tsv',
        reference: str = 'GRCh38', pca_type: str = None,
        data_dirname: str = None, data_basename: str = None, input_type: str = None,
        maf: float = 0.05, hwe: float = 1e-3, call_rate: float = 0.98,
        ld_cor: float = 0.2, ld_window: int = 250000, n_pcs: int = 20, relatedness_method: str = 'pc_relate',
        relatedness_thresh: float = 0.98, prob_threshold: float = 0.8, out_dir: str = None):

    if not out_dir:
        raise Exception('\nOutput directory where files will be saved is not specified')

    if pca_type == 'project':
        print('\nRunning PCA using projection method')

        from gwaspy.pca.pca_project import run_pca_project
        run_pca_project(ref_dirname=ref_dirname, ref_basename=ref_basename, ref_info=ref_info,
                        data_dirname=data_dirname, data_basename=data_basename, out_dir=out_dir, input_type=input_type,
                        reference=reference, npcs=n_pcs, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window,
                        prob_threshold=prob_threshold)

    elif pca_type == 'joint':
        print('\nRunning PCA using joint method')
        from gwaspy.pca.pca_joint import run_pca_joint
        run_pca_joint(ref_dirname=ref_dirname, ref_basename=ref_basename, ref_info=ref_info, data_dirname=data_dirname,
                      data_basename=data_basename, out_dir=out_dir, input_type=input_type, reference=reference,
                      npcs=n_pcs, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window,
                      relatedness_method=relatedness_method, relatedness_thresh=relatedness_thresh,
                      prob_threshold=prob_threshold)

    else:
        print('\nRunning PCA without a reference')
        from gwaspy.pca.pca_normal import run_pca_normal
        run_pca_normal(dirname=data_dirname, basename=data_basename, input_type=input_type, out_dir=out_dir,
                       reference=reference, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window,
                       n_pcs=n_pcs, relatedness_method=relatedness_method, relatedness_thresh=relatedness_thresh)


def main():
    parser = argparse.ArgumentParser()
    # reference args
    parser.add_argument('--ref-dirname', default='gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/')
    parser.add_argument('--ref-basename', default='unrelated')
    parser.add_argument('--ref-info', default='gs://hgdp-1kg/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_sample_info.unrelateds.pca_outliers_removed.with_project.tsv')
    parser.add_argument('--reference', type=str, default='GRCh38')
    parser.add_argument('--pca-type', type=str, default='normal', choices=['normal', 'project', 'joint'])

    # data args
    parser.add_argument('--data-dirname', type=str, required=True)
    parser.add_argument('--data-basename', type=str, required=True)
    parser.add_argument('--input-type', type=str, required=True, choices=['vcf', 'plink', 'hail'])

    # filter args
    parser.add_argument('--maf', type=float, default=0.05, help='include only SNPs with MAF >= NUM in PCA')
    parser.add_argument('--hwe', type=float, default=1e-3, help='include only SNPs with HWE >= NUM in PCA')
    parser.add_argument('--geno', type=float, default=0.98, help='include only SNPs with call-rate > NUM')
    parser.add_argument('--ld-cor', type=float, default=0.2, choices=range(0,1), metavar="[0.0-1.0]",
                        help='Squared correlation threshold (exclusive upper bound). Must be in the range [0.0, 1.0]')
    parser.add_argument('--ld-window', type=int, default=250000,
                        help='Window size in base pairs (inclusive upper bound)')
    parser.add_argument('--npcs', type=int, default=20, help='Number of PCs to use')
    parser.add_argument('--relatedness-method', type=str, default='pc_relate',
                        choices=['pc_relate', 'ibd', 'king'], help='Method to use for the inference of relatedness')
    parser.add_argument('--relatedness-thresh', type=float, default=0.98,
                        help='Threshold value to use in relatedness checks')
    parser.add_argument('--prob', type=float, default=0.8,
                        help='Minimum probability of belonging to a given population for the population to be set')
    parser.add_argument('--out-dir', type=str, required=True)

    args = parser.parse_args()

    if not args.prob:
        print(f'No prob value specified, {args.prob} will be used')

    hl.init(default_reference=args.reference)

    pca(ref_dirname=args.ref_dirname, ref_basename=args.ref_basename, ref_info=args.ref_info, reference=args.reference,
        pca_type=args.pca_type, input_type=args.input_type, data_dirname=args.data_dirname,
        data_basename=args.data_basename, maf=args.maf, hwe=args.hwe, call_rate=args.geno, ld_cor=args.ld_cor,
        ld_window=args.ld_window, n_pcs=args.npcs, relatedness_method=args.relatedness_method,
        relatedness_thresh=args.relatedness_thresh, prob_threshold=args.prob, out_dir=args.out_dir)

    print('\nDone running PCA')


if __name__ == '__main__':
    main()




__author__ = 'Lindo Nkambule'

import hail as hl
import pandas as pd
from gwaspy.pca.pca_filter_snps import pca_filter_mt, relatedness_check
import plotly.express as px


def joint_pca(
        ref_dirname: str = 'gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/',
        ref_basename: str = 'unrelated',
        in_mt: hl.MatrixTable = None,
        data_basename: str = None,
        npcs: int = 20,
        out_dir: str = None):
    """
    Merges input dataset with ref by [locus, alleles] and runs PCA on merged dataset
    :param ref_dirname: directory name where reference data is
    :param ref_basename: base filename for reference data
    :param in_mt: input data MatrixTable
    :param data_basename: base filename for input data
    :param npcs: number of principal components to be used in PCA
    :param out_dir: output directory where files are going to be saved to
    :return:
    """
    print('\nReading reference data mt')
    ref_mt = hl.read_matrix_table(f'{ref_dirname}{ref_basename}.mt')

    # We need to unkey the datasets and take only cols common between the two in order to be able to merge in Hail
    ref_mt = ref_mt.key_cols_by().key_rows_by()
    ref_downsampled = ref_mt.select_cols('s').select_rows('locus', 'alleles').select_entries('GT')
    ref_downsampled = ref_downsampled.key_cols_by('s').key_rows_by('locus', 'alleles')

    data_mt = in_mt.key_cols_by().key_rows_by()
    data_downsampled = data_mt.select_cols('s').select_rows('locus', 'alleles').select_entries('GT')
    data_downsampled = data_downsampled.key_cols_by('s').key_rows_by('locus', 'alleles')

    print('\nJoining Data with Ref by locus & alleles')
    joined = ref_downsampled.union_cols(data_downsampled)

    pca_snps = joined.count_rows()
    if pca_snps > 1000000:
        import warnings
        warnings.warn(f'Too many SNPs to be used in PCA: {pca_snps}. This will make PCA run longer')

    print(f'\nRunning PCA with {npcs} principal components')
    pca_evals, pca_scores, _ = hl.hwe_normalized_pca(joined.GT, k=npcs)

    pca_scores = pca_scores.transmute(**{f'PC{i}': pca_scores.scores[i - 1] for i in range(1, npcs+1)})
    print(f'\nExporting PCA scores to {out_dir}')
    pca_scores.export(f'{out_dir}GWASpy/PCA/{data_basename}/pca_joint/{data_basename}.1kg_hgdp.joint.pca.scores.txt.bgz')


def add_ref_superpop_labels(joint_scores: str = None, ref_info: str = None) -> pd.DataFrame:
    """
    Add SuperPop labels to reference samples so we can use them in the RF model to assign ancestry
    :param joint_scores: path to joint data+ref scores
    :param ref_info: path to information about samples in the ref scores
    :return: a pandas Dataframe of joint data+ref scores with SuperPop label for ref samples
    """
    print('\nAdding SuperPop labels to reference samples')
    joint_data = pd.read_table(joint_scores, header=0, sep='\t', compression='gzip')
    ref_info = pd.read_table(ref_info, header=0, sep='\t')
    updated_joint_data = pd.merge(left=joint_data, right=ref_info, left_on='s', right_on='Sample', how='left')

    return updated_joint_data


def plot_pca_joint(joint_scores: pd.DataFrame = None, x_pc: str = None, y_pc: str = None):
    """
    Plot PCs
    :param joint_scores: pandas dataframe of joint data+ref scores
    :param x_pc: x-axis pc
    :param y_pc: y-axis pc
    :return: pc plot
    """
    # split data and ref into separate dfs
    data = joint_scores.loc[joint_scores['SuperPop'].isnull()]
    ref = joint_scores.loc[joint_scores['SuperPop'].notnull()]

    data['Project'] = "Input Dataset"
    data = data[['s', 'pop', 'Project', x_pc, y_pc]]
    ref = ref[['s', 'SuperPop', 'Project', x_pc, y_pc]]
    ref.rename(columns={'SuperPop': 'pop'}, inplace=True)

    # concatenate the two dfs together
    concatenated = pd.concat([ref, data], axis=0)

    fig = px.scatter(concatenated, x=x_pc, y=y_pc, color='pop',
                     hover_data=['s', x_pc, y_pc, 'pop', 'Project'],
                     color_discrete_map={'AFR': "#984EA3", 'EAS': "#4DAF4A", 'EUR': "#377EB8", 'CSA': "#FF7F00",
                                         'AMR': "#E41A1C", 'MID': "#A65628", 'OCE': "#999999",
                                         'oth': "#F0E442"}).update_traces(marker_size=4)

    return fig


def run_pca_joint(
        ref_dirname: str = 'gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/',
        ref_basename: str = 'unrelated',
        ref_info: str = 'gs://hgdp-1kg/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_sample_info.unrelateds.pca_outliers_removed.with_project.tsv',
        data_dirname: str = None,
        data_basename: str = None,
        out_dir: str = None,
        input_type: str = None,
        reference: str = 'GRCh38',
        npcs: int = 20,
        maf: float = 0.05,
        hwe: float = 1e-3,
        call_rate: float = 0.98,
        ld_cor: float = 0.2,
        ld_window: int = 250000,
        relatedness_method: str = 'pc_relate',
        relatedness_thresh: float = 0.98,
        prob_threshold: float = 0.8):
    """
    Project samples into predefined PCA space
    :param ref_dirname: directory name where reference data is
    :param ref_basename: base filename for reference data
    :param ref_info: reference sample information
    :param data_dirname: matrix table of data to project
    :param data_basename: matrix table of data to project
    :param out_dir: directory and filename prefix for where to put PCA projection output
    :param input_type: input file(s) type: hail, plink, or vcf
    :param reference: reference build
    :param npcs: number of principal components to be used in PCA
    :param maf: minor allele frequency threshold
    :param hwe: hardy-weinberg filter threshold
    :param call_rate: variant call rate filter threshold
    :param ld_cor: reference build
    :param ld_window: window size
    :param prob_threshold: a list of probability thresholds to use for classifying samples
    :param relatedness_method: method to use for relatedness filtering
    :param relatedness_thresh: threshold to use for filtering out related individuals
    :return: a pandas Dataframe with data PCA scores projected on the same PCA space using the Human Genome Diversity
    """
    print('\nReading data mt')
    if reference.lower() == 'grch37':
        lifted_over = f'{data_dirname}{data_basename}.liftover.grch38.mt'
        if not hl.hadoop_exists(lifted_over):
            from gwaspy.utils.reference_liftover import liftover_to_grch38
            mt = liftover_to_grch38(dirname=data_dirname, basename=data_basename, input_type=input_type)
        else:
            print(f'\nFound lifted-over over file: {lifted_over}')
            mt = hl.read_matrix_table(lifted_over)
    else:
        from gwaspy.utils.read_file import read_infile
        mt = read_infile(input_type=input_type, dirname=data_dirname, basename=data_basename)

    print("\nFiltering data mt")
    data_mt = pca_filter_mt(in_mt=mt, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window)

    data_mt = relatedness_check(in_mt=data_mt, method=relatedness_method, outdir=out_dir,
                                kin_estimate=relatedness_thresh)

    joint_pca(ref_dirname=ref_dirname, ref_basename=ref_basename, in_mt=data_mt, data_basename=data_basename, npcs=npcs,
              out_dir=out_dir)

    scores_without_pop_label = f'{out_dir}GWASpy/PCA/{data_basename}/pca_joint/{data_basename}.1kg_hgdp.joint.pca.scores.txt.bgz'
    scores_with_pop_label_df = add_ref_superpop_labels(joint_scores=scores_without_pop_label, ref_info=ref_info)

    from gwaspy.pca.assign_pop_labels import assign_population_pcs
    pcs_df, clf = assign_population_pcs(pop_pc_pd=scores_with_pop_label_df, num_pcs=npcs, min_prob=prob_threshold)

    data_pops = pcs_df.loc[pcs_df['SuperPop'].isnull()]
    data_pops['pop'].value_counts()
    cols = ['s', 'pop'] + [f'prob_{i}' for i in ["AFR", "AMR", "CSA", "EAS", "EUR", "MID", "OCE"]] + [f'PC{i}' for i in
                                                                                                      range(1, npcs+1)]
    data_pops_df = data_pops[cols]

    data_pops_df.to_csv(f'{out_dir}GWASpy/PCA/{data_basename}/pca_joint/pca_sup_pops_{prob_threshold}_probs.joint.pca.txt',
                        sep='\t', index=False)

    print('\nGenerating PCA plots')

    figs_dict = {}
    # plotting more than 10 PCA plots in HTML generates wobbly, large files
    for i in range(1, 10, 2):
        xpc = f'PC{i}'
        ypc = f'PC{i + 1}'
        figs_dict["fig{}{}".format(xpc, ypc)] = plot_pca_joint(joint_scores=pcs_df,
                                                               x_pc=xpc, y_pc=ypc)

    with open('/tmp/joint.pca.plots.html', 'a') as f:
        for figname, figure in figs_dict.items():
            f.write(figure.to_html(include_plotlyjs='cdn'))

    hl.hadoop_copy('file:///tmp/joint.pca.plots.html',
                   f'{out_dir}GWASpy/PCA/{data_basename}/pca_joint/{data_basename}.joint.pca.plots.html')




__author__ = 'Lindo Nkambule'

from gwaspy.preimp_qc.annotations import *
from typing import Tuple, Any, Dict, Union
from gwaspy.utils.read_file import read_infile
import argparse
from gwaspy.preimp_qc.report import MyDocument
import shutil
import warnings
import os
import hail as hl

warnings.simplefilter(action='ignore', category=RuntimeWarning)


def summary_stats(mt: hl.MatrixTable) -> Tuple[hl.MatrixTable, Dict[str, Any]]:
    results = {}

    if 'is_case' in mt.col:
        counts = mt.aggregate_cols(hl.struct(is_case=hl.agg.counter(mt.is_case),
                                             is_female=hl.agg.counter(mt.is_female)))

        mapping = {
            False: 'control',
            True: 'case',
            None: 'unknown'
        }
        is_case_counts = {mapping[pheno]: count for pheno, count in counts['is_case'].items()}
        results['is_case_counts'] = is_case_counts

    else:
        counts = mt.aggregate_cols(hl.struct(is_female=hl.agg.counter(mt.is_female)))

    mapping = {
        True: 'female',
        False: 'male',
        None: 'unknown'
    }
    is_female_counts = {mapping[pheno]: count for pheno, count in counts['is_female'].items()}
    results['is_female_counts'] = is_female_counts

    if 'is_case' in mt.col:
        pheno_status = ['case', 'control', 'unknown']

        for i in pheno_status:
            if i not in results['is_case_counts']:
                results['is_case_counts'][i] = 0

    sex_status = ['female', 'male', 'unknown']

    for i in sex_status:
        if i not in results['is_female_counts']:
            results['is_female_counts'][i] = 0

    n_variants = mt.count_rows()
    n_samples = mt.count_cols()

    results['n_variants'] = n_variants
    results['n_samples'] = n_samples

    return mt, results


def preimp_qc(input_type: str = None, dirname: str = None, basename: str = None, pre_geno_thresh: Union[int, float] = 0.95,
              mind_thresh: Union[int, float] = 0.98, fhet_aut: Union[int, float] = 0.2, fstat_x: Union[int, float] = 0.5,
              fstat_y: Union[int, float] = 0.5, geno_thresh: Union[int, float] = 0.98,
              cr_diff_thresh: Union[int, float] = 0.02, maf_thresh: Union[int, float] = 0.01,
              hwe_th_con_thresh: Union[int, float] = 1e-6, hwe_th_cas_thresh: Union[int, float] = 1e-10,
              hwe_th_all_thresh: Union[int, float] = 1e-06, annotations_file: str = None, report: bool = True,
              export_type: str = 'hail', out_dir: str = None, reference: str = 'GRCh38'):
    print('\nRunning QC')

    global mt, row_filters, filters, data_type, lambda_gc_pos, lambda_gc_pre, n_sig_var_pre, n_sig_var_pos, man_table_results, remove_fields

    # create temp directory for storing temp files
    if not os.path.exists('gwaspy_tmp'):
        os.makedirs('gwaspy_tmp')

    # LaTex needs full path to files
    gwaspy_dir = os.getcwd() + '/gwaspy_tmp'

    output_directory = out_dir if out_dir else dirname

    hl.init(default_reference=reference)

    # read input
    mt = read_infile(input_type=input_type, dirname=dirname, basename=basename, annotations=annotations_file)

    if 'is_case' in mt.col:
        gwas_pre, n_sig_var_pre = manhattan(qqtitle='Pre-QC QQ Plot', mantitle='Pre-QC Manhattan Plot').filter(mt)
        qqplt_pre, lambda_gc_pre, manplt_pre = manhattan(qqtitle='Pre-QC QQ Plot',
                                                         mantitle='Pre-QC Manhattan Plot').plot(gwas_pre)
        qqplt_pre.savefig('gwaspy_tmp/gwaspy_qq_pre.png', dpi=300)
        manplt_pre.savefig('gwaspy_tmp/gwaspy_man_pre.png', dpi=300)

    mt = mt.annotate_rows(exclude_row=False)
    mt = mt.annotate_cols(exclude_col=False)

    mt, pre_qc_counts = summary_stats(mt)

    if 'is_case' in mt.col:
        if (pre_qc_counts['is_case_counts']['case'] > 0) & (pre_qc_counts['is_case_counts']['control'] == 0):
            data_type = 'Case-only'
        elif (pre_qc_counts['is_case_counts']['control'] > 0) & (pre_qc_counts['is_case_counts']['case'] == 0):
            data_type = 'Control-only'
        elif (pre_qc_counts['is_case_counts']['case'] > 0) & (pre_qc_counts['is_case_counts']['control'] > 0):
            data_type = 'Case-Control'
        else:
            data_type = 'Trio'
    else:
        data_type = 'no-pheno'

    chroms = mt.aggregate_rows(hl.agg.collect_as_set(mt.locus.contig))
    if ('chrX' or 'chrY' or 'chrMT') in chroms:
        chromx, chromy, chrommt = 'chrX', 'chrY', 'chrMT'
    else:
        chromx, chromy, chrommt = 'X', 'Y', 'MT'

    # we need to compute call rate for chr1-23 and chrY separately since females have no chrY
    mt = mt.annotate_entries(
        geno_y_excluded=(hl.case()
                         .when(mt.locus.contig == chromy, False)
                         .default(True)
                         ),
        geno_y_only=(hl.case()
                     .when(mt.locus.contig == chromy, mt.is_female == False)
                     .default(False)
                     )
    )

    mt = pre_geno(pre_geno_cr=pre_geno_thresh).filter(mt)
    mt = id_call_rate(mind=mind_thresh, pre_row_filter='pre_geno').filter(mt)

    mt = fhet_autosomes(pre_row_filter='pre_geno', fhet_thresh=fhet_aut).filter(mt)
    mt = fhet_sex(pre_row_filter='pre_geno', fstat_x=fstat_x, fstat_y=fstat_y).filter(mt)
    mt = fhet_sex_warnings(pre_row_filter='pre_geno', pre_col_filter='sex_violations').filter(mt)

    mt = mt.annotate_cols(**{
        'id_pass': hl.struct(
            filters=((hl.agg.any(mt['mind'].filters) == True) | (hl.agg.any(mt['fstat'].filters) == True) |
                     (hl.agg.any(mt['sex_violations'].filters) == True))
        )})

    mt = geno(pre_row_filter='pre_geno', pre_col_filter='id_pass', geno_thresh=geno_thresh, data_type=data_type).filter(mt)
    mt = invariant(pre_col_filter='id_pass').filter(mt)

    # for HWE, markers in: (1) autosomes - include males+females; (2) chrX - include ONLY females; (3) exclude chrY
    mt = mt.annotate_entries(
        hwe_aut=(hl.case()
                 .when(mt.locus.contig == chromx, False)
                 .when(mt.locus.contig == chromy, False)
                 .when(mt.locus.contig == chrommt, False)
                 .default(True)
                 ),
        hwe_sex=(hl.case()
                 .when(mt.locus.contig == chromx, mt.is_female)
                 .default(False)
                 )
    )

    if 'is_case' in mt.col:
        mt = call_rate_diff(pre_row_filter='geno', pre_col_filter='id_pass', initial_row_filter='pre_geno',
                            cr_thresh=cr_diff_thresh).filter(mt)

        # check if data is case-/control-only, case-control, or trio
        # (a) Case-Only
        if data_type == 'Case-only':
            print("\n" + data_type)
            mt = hwe_cas(pre_col_filter='id_pass', pre_row_filter='geno', hwe_th_ca=1e-6).filter(mt)
            row_filters = ['pre_geno', 'geno', 'cr_diff', 'monomorphic_var', 'hwe_cas']
            filters = ['pre_geno', 'mind', 'fstat', 'sex_violations', 'sex_warnings', 'geno', 'cr_diff',
                       'monomorphic_var', 'hwe_cas']
            remove_fields = ['cr', 'diff', 'hwe_cas_aut', 'hwe_cas_sex']
        # (b) Control-Only
        elif data_type == 'Control-only':
            print("\n" + data_type)
            mt = hwe_con(pre_col_filter='id_pass', pre_row_filter='geno', hwe_th_co=1e-6).filter(mt)
            row_filters = ['pre_geno', 'geno', 'cr_diff', 'monomorphic_var', 'hwe_con']
            filters = ['pre_geno', 'mind', 'fstat', 'sex_violations', 'sex_warnings', 'geno', 'cr_diff',
                       'monomorphic_var', 'hwe_con']
            remove_fields = ['cr', 'diff', 'hwe_con_aut', 'hwe_con_sex']
        elif data_type == 'Case-Control':
            print("\n" + data_type)
            mt = hwe_cas(pre_col_filter='id_pass', pre_row_filter='geno', hwe_th_ca=hwe_th_cas_thresh).filter(mt)
            mt = hwe_con(pre_col_filter='id_pass', pre_row_filter='geno', hwe_th_co=hwe_th_con_thresh).filter(mt)
            row_filters = ['pre_geno', 'geno', 'cr_diff', 'monomorphic_var', 'hwe_con', 'hwe_cas']
            filters = ['pre_geno', 'mind', 'fstat', 'sex_violations', 'sex_warnings', 'geno', 'cr_diff',
                       'monomorphic_var', 'hwe_con', 'hwe_cas']
            remove_fields = ['cr', 'diff', 'hwe_cas_aut', 'hwe_cas_sex', 'hwe_con_aut', 'hwe_con_sex']
        else:
            # trio data
            print(data_type)
    else:
        print('Running HWE filters on whole dataset without spliting by phenotype status')
        mt = hwe_all(pre_col_filter='id_pass', pre_row_filter='geno', hwe_th_all=1e-08).filter(mt)
        row_filters = ['pre_geno', 'geno', 'monomorphic_var', 'hwe_all']
        filters = ['pre_geno', 'mind', 'fstat', 'sex_violations', 'sex_warnings', 'geno',
                   'monomorphic_var', 'hwe_all']
        remove_fields = ['hwe_all_aut', 'hwe_all_sex']

    results = {}
    column_filters = ['mind', 'fstat', 'sex_violations', 'sex_warnings']

    mt.select_entries().select_rows(*row_filters).select_cols(*column_filters).write(f'{output_directory}/temp.mt',
                                                                                     overwrite=True)
    mt_temp = hl.read_matrix_table(f'{output_directory}/temp.mt')
    column_aggregations = mt_temp.aggregate_cols(
        [hl.agg.counter(mt_temp[filter].filters) for filter in column_filters])

    row_aggregations = mt_temp.aggregate_rows(
        [hl.agg.counter(mt_temp[filter].filters) for filter in row_filters])

    for filt, cont in zip(column_filters, column_aggregations):
        results[filt] = dict(cont) # aggregate returns a frozendict, convert that back to a dict

    for filt, cont in zip(row_filters, row_aggregations):
        results[filt] = dict(cont) # # aggregate returns a frozendict, convert that back to a dict

    for i in filters:
        # some filters will have zero snps/id filtered, and there won't be a True, so add it
        if True not in results[i]:
            results[i][True] = 0
        for key, value in results.items():
            if i == key:
                print(key, ': ', value)

    if report:
        fstat_fig = fhet_sex(pre_row_filter='pre_geno', fstat_x=fstat_x, fstat_y=fstat_y, figsize=(15, 20)).plot(mt)
        fstat_fig.savefig('gwaspy_tmp/gwaspy_fstat_fig.png', dpi=300)

        id_cr_plot = id_call_rate(mind=mind_thresh, pre_row_filter='pre_geno', data_type=data_type).plot(mt)
        var_cr_plot = geno(pre_row_filter='pre_geno', pre_col_filter='id_pass', geno_thresh=geno_thresh,
                           data_type=data_type).plot(mt)

        if 'is_case' in mt.col:
            if data_type == 'Case-only':
                id_cr_plot[0].savefig('gwaspy_tmp/gwaspy_id_cas_pre.png', dpi=300)
                var_cr_plot[0].savefig('gwaspy_tmp/gwaspy_var_cas_pre.png', dpi=300)

            if data_type == 'Control-only':
                id_cr_plot[0].savefig('gwaspy_tmp/gwaspy_id_con_pre.png', dpi=300)
                var_cr_plot[0].savefig('gwaspy_tmp/gwaspy_var_con_pre.png', dpi=300)

            if data_type == 'Case-Control':
                id_cr_plot[0].savefig('gwaspy_tmp/gwaspy_id_con_pre.png', dpi=300)
                id_cr_plot[1].savefig('gwaspy_tmp/gwaspy_id_cas_pre.png', dpi=300)
                var_cr_plot[0].savefig('gwaspy_tmp/gwaspy_var_con_pre.png', dpi=300)
                var_cr_plot[1].savefig('gwaspy_tmp/gwaspy_var_cas_pre.png', dpi=300)

        else:
            id_cr_plot[0].savefig('gwaspy_tmp/gwaspy_id_cas_con_pre.png', dpi=300)
            var_cr_plot[0].savefig('gwaspy_tmp/gwaspy_var_cas_con_pre.png', dpi=300)

    # FILTER OUT ALL SNPs and IDs THAT FAIL QC
    column_filters = ['mind', 'fstat', 'sex_violations']
    for row in row_filters:
        mt = mt.filter_rows(mt[row].filters == True, keep=False)
    for col in column_filters:
        mt = mt.filter_cols(mt[col].filters == True, keep=False)

    mt_filtered, pos_qc_counts = summary_stats(mt)

    # drop entry fields we added as they will cause errors when exporting to VCF
    # e.g. Error summary: HailException: Invalid type for format field 'geno_y_excluded'. Found 'bool'.
    drop_fields = filters + ['geno_y_excluded', 'geno_y_only', 'pre_geno_noy', 'pre_geno_y', 'hwe_aut', 'hwe_sex',
                             'exclude_col', 'exclude_row', 'variant_qc', 'aaf', 'geno_noy', 'geno_y', 'sex_ambiguous',
                             'id_pass'] + remove_fields
    mt_filtered = mt_filtered.drop(*drop_fields)

    if 'is_case' in mt.col:
        gwas_pos, n_sig_var_pos = manhattan(qqtitle='Post-QC QQ Plot', mantitle='Post-QC Manhattan Plot').filter(mt)
        qqplt_pos, lambda_gc_pos, manplt_pos = manhattan(qqtitle='Post-QC QQ Plot',
                                                         mantitle='Post-QC Manhattan Plot').plot(gwas_pos)

        ncas_pre = pre_qc_counts['is_case_counts']['case']
        ncas_pos = pos_qc_counts['is_case_counts']['case']
        ncon_pre = pre_qc_counts['is_case_counts']['control']
        ncon_pos = pos_qc_counts['is_case_counts']['control']
        lambda_thous_pre = 1 + (lambda_gc_pre-1)*(1/ncas_pre+1/ncon_pre)/(1/1000+1/1000)
        lambda_thous_pos = 1 + (lambda_gc_pos-1)*(1/ncas_pos+1/ncon_pos)/(1/1000+1/1000)

        qqplt_pos.savefig('gwaspy_tmp/gwaspy_qq_pos.png', dpi=300)
        manplt_pos.savefig('gwaspy_tmp/gwaspy_man_pos.png', dpi=300)

        man_table_results = [n_sig_var_pre, n_sig_var_pos, lambda_gc_pre, lambda_gc_pos, round(lambda_thous_pre, 3),
                             round(lambda_thous_pos, 3)]

    # report
    if report:
        print('\nWriting report')
        doc = MyDocument(basename=basename)
        if 'is_case' in mt.col:
            doc.flags_table(pre_qc_counts=pre_qc_counts, pos_qc_counts=pos_qc_counts, results=results,
                            lambda_gc=lambda_gc_pos, sig_vars=n_sig_var_pos)
        else:
            doc.flags_table(pre_qc_counts=pre_qc_counts, pos_qc_counts=pos_qc_counts, results=results)
        doc.general_info(pre_qc_conts=pre_qc_counts, post_qc_conts=pos_qc_counts,
                         count_results=results, pre_filter=pre_geno_thresh, id_cr=mind_thresh, fhet_thresh=fhet_aut,
                         var_cr=geno_thresh, miss_diff=cr_diff_thresh, hwe_con=hwe_th_con_thresh,
                         hwe_cas=hwe_th_cas_thresh, hwe_all=hwe_th_all_thresh, data_type=data_type)
        if 'is_case' in mt.col:
            doc.manhattan_sec(qq_pre_path=f'{gwaspy_dir}/gwaspy_qq_pre.png', qq_pos_path=f'{gwaspy_dir}/gwaspy_qq_pos.png',
                              man_pre_path=f'{gwaspy_dir}/gwaspy_man_pre.png',
                              man_pos_path=f'{gwaspy_dir}/gwaspy_man_pos.png',
                              table_results=man_table_results)

        if data_type == 'Case-only':
            doc.individual_char(id_con_pre_path='nothing here', id_cas_pre_path=f'{gwaspy_dir}/gwaspy_id_cas_pre.png',
                                fstat_fig_path=f'{gwaspy_dir}/gwaspy_fstat_fig.png', id_all_path='nothing here',
                                data_type=data_type)
            doc.snp_char(var_con_pre_path='nothing here', var_cas_pre_path=f'{gwaspy_dir}/gwaspy_var_cas_pre.png',
                         var_all_path='nothing here', data_type=data_type)
        if data_type == 'Control-only':
            doc.individual_char(id_con_pre_path=f'{gwaspy_dir}/gwaspy_id_con_pre.png', id_cas_pre_path='nothing here',
                                fstat_fig_path=f'{gwaspy_dir}/gwaspy_fstat_fig.png', id_all_path='nothing here',
                                data_type=data_type)
            doc.snp_char(var_con_pre_path=f'{gwaspy_dir}/gwaspy_var_cas_pre.png', var_cas_pre_path='nothing here',
                         var_all_path='nothing here', data_type=data_type)
        if data_type == 'Case-Control':
            doc.individual_char(id_con_pre_path=f'{gwaspy_dir}/gwaspy_id_con_pre.png',
                                id_cas_pre_path=f'{gwaspy_dir}/gwaspy_id_cas_pre.png',
                                fstat_fig_path=f'{gwaspy_dir}/gwaspy_fstat_fig.png', id_all_path='nothing here',
                                data_type=data_type)
            doc.snp_char(var_con_pre_path=f'{gwaspy_dir}/gwaspy_var_cas_pre.png',
                         var_cas_pre_path=f'{gwaspy_dir}/gwaspy_var_cas_pre.png',
                         var_all_path='nothing here', data_type=data_type)
        if data_type == 'no-pheno':
            doc.individual_char(id_con_pre_path='nothing here', id_cas_pre_path='nothing here',
                                fstat_fig_path=f'{gwaspy_dir}/gwaspy_fstat_fig.png',
                                id_all_path=f'{gwaspy_dir}/gwaspy_id_cas_con_pre.png', data_type=data_type)
            doc.snp_char(var_con_pre_path='nothing here', var_cas_pre_path=f'nothing here',
                         var_all_path=f'{gwaspy_dir}/gwaspy_var_cas_con_pre.png', data_type=data_type)
        doc.generate_pdf(f'{gwaspy_dir}/{basename}.preimp_qc.report', clean=True, clean_tex=True)

    print('\nExporting qced file')
    if export_type:
        from gwaspy.utils.export_file import export_qced_file
        export_qced_file(mt=mt_filtered, out_dir=output_directory, basename=basename, export_type=export_type)

    if output_directory.startswith('gs://'):
        hl.hadoop_copy(f'file://{gwaspy_dir}/{basename}.preimp_qc.report.pdf',
                       f'{output_directory}GWASpy/Preimp_QC/{basename}.preimp_qc.report.pdf')
    else:
        shutil.copyfile(f'{gwaspy_dir}/{basename}.preimp_qc.report.pdf',
                        f'{output_directory}{basename}.preimp_qc.report.pdf')

    # clean-up
    print('\nCleaning up')
    shutil.rmtree('gwaspy_tmp')

    print("\nDone running QC!")


def main():
    parser = argparse.ArgumentParser(description='preimp_qc')
    parser.add_argument('--dirname', type=str, required=True)
    parser.add_argument('--basename', type=str, required=True)
    parser.add_argument('--input-type', type=str, required=True, choices=['vcf', 'plink', 'hail'])
    parser.add_argument('--export-type', type=str, default='hail', choices=['vcf', 'plink', 'hail'])
    parser.add_argument('--out-dir', type=str, default=None)
    parser.add_argument('--annotations', type=str)
    parser.add_argument('--reference', type=str, default='GRCh38')
    parser.add_argument('--report', action='store_false')
    # parser.add_argument('--qc_round', type=str, required=True)

    # required for QC
    parser.add_argument('--pre-geno', type=float, default=0.95,
                        help="include only SNPs with missing-rate < NUM (before "
                             "ID filter), important for post merge of multiple "
                             "platforms")
    parser.add_argument('--mind', type=float, default=0.98, help="include only IDs with missing-rate < NUM")
    parser.add_argument('--fhet-aut', type=float, default=0.2, help="include only IDs within NUM < FHET < NUM")
    parser.add_argument('--fstat-y', type=float, default=0.5, help="include only female IDs with fhet < NUM")
    parser.add_argument('--fstat-x', type=float, default=0.5, help="include only male IDs with fhet > NUM")
    parser.add_argument('--geno', type=float, default=0.98, help="include only SNPs with missing-rate > NUM")
    parser.add_argument('--midi', type=float, default=0.02, help="include only SNPs with missing-rate-difference ("
                                                                 "case/control) < NUM")
    parser.add_argument('--withpna', type=int, default=0, choices=[0, 1], help="include monomorphic (invariant) SNPs")
    parser.add_argument('--maf', type=float, default=0.01, help="include only SNPs with MAF >= NUM")
    parser.add_argument('--hwe-th-con', type=float, default=1e-06, help="HWE controls < NUM")
    parser.add_argument('--hwe-th-cas', type=float, default=1e-10, help="HWE cases < NUM")
    parser.add_argument('--hwe-th-all', type=float, default=1e-06, help="HWE cases + controls < NUM")

    arg = parser.parse_args()

    preimp_qc(input_type=arg.input_type, dirname=arg.dirname, basename=arg.basename, pre_geno_thresh=arg.pre_geno,
              mind_thresh=arg.mind, fhet_aut=arg.fhet_aut, fstat_x=arg.fstat_x, fstat_y=arg.fstat_y,
              geno_thresh=arg.geno, cr_diff_thresh=arg.midi, maf_thresh=arg.maf, hwe_th_con_thresh=arg.hwe_th_con,
              hwe_th_cas_thresh=arg.hwe_th_cas, hwe_th_all_thresh=arg.hwe_th_all, annotations_file=arg.annotations,
              report=arg.report, export_type=arg.export_type, out_dir=arg.out_dir, reference=arg.reference)


if __name__ == '__main__':
    main()

